// sim_hol.cc (c) Matt Clark Septemeber 2002
// This is released under GPL - see the file COPYING for
// full details.
// 
// This program uses the public available (GPL) fftw routines for computing
// FFTs- you must either obtain FFTW from www.fftw.org or replace the FFT
// function calls with your own FFT routines - I recommend FFTW.  If you are 
// lucky FFTW may have been bundled with this software.
//

/*  This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */




#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <complex>
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fftw3.h>
using namespace std;
#ifndef WISDOM_FILE
#define WISDOM_FILE "/tmp/sim_hol.wisdom"
#endif

#define PI      M_PI
#define FALSE   0
#define TRUE    1
#define ST strcmp
#define TS temp_string
#define BOOL    int
#define MAX_OPTIONS 2500
#define MAX_S	1024
 

//prototypes for functions
void 		hol_to_sim(char *, int, int, float, complex <double> *, int, float, float, int);
void 		deal_with_plan(int N, char *,fftw_plan * ,fftw_plan *);	
void 		add_wavefront(complex<double> *, int, int, int, float, float, float);
void            intext_gaussian(char *,complex<double> *, int , int ,int ,int ,int ,float );
void            intext  (char *, complex<double> *, int,int, int, int, int);
float           intensity(complex<double>);
void 		prop2d(complex<double> * , float ,float ,float ,int);
inline complex<double> 	rot(complex<double> , double );

//---global variable for the cray fft routine workspace stuff---------

main(int argc,char **argv){
        ifstream        control;
	FILE		*fo,*wisd_file;
        complex<double>	*here,*there;
	char 		*hologram;
        int             phaselevels=0,i,N=0,nx=0,ny=0;
        float           wavelength=633e-9,F=0;
	float		psize,F_bar=0,F_dash=0;
        char            control_file[MAX_S];
        char            iname[MAX_S];
	char		oname[MAX_S];
	char		temp_string[MAX_S];
	BOOL		autoscale=FALSE;
// FFTW
	fftw_plan forward,backward;

cout<<"Sim_hol- (c) matt clark\n";
cout<<"This program is released under GPL\n";
cout<<"Part of the sim_wf suite of programs\n";
cout<<"Based on the sim_ap skeleton v 4.30\n";


// first check and open control file
// start of the program
         if (argc > 1) {
                strcpy(control_file, argv[1]);
                }
        else {
                cout<<"SIM_HOL: Error: usage sim_hol controlfile\n";
		cout<<"In the control file the following commands are support\n\n";
		cout<<"Absolutely required\n";
		cout<<"inname  		<hologramname>\n";
		cout<<"outname 		<reconstruction_name>\n";
		cout<<"focal_length    	<F>\n";
		cout<<"psize   		<px>\n";
		cout<<"pixels  		<nx ny>\n";
		cout<<"phase 		<px>\n\n";
		cout<<"Either\n";
		cout<<"autoscale (and optionally N)\n";
		cout<<"or\n";
		cout<<"N		<N> (and optionally input_curvature)\n";
		exit(1);
                }
        control.open(control_file);
        if(control.bad()){
		cout<<"ERROR:: unable to open control file "<<control_file<<"\n";
                exit(3);
                }
// readin file and find 'options' bit and then set options
        i=0;
	do{
		i++;
                control>>temp_string;
	        if(ST(TS,"input_curvature")==0){
			control>>F_bar;
			}
		if(ST(TS,"wavelength")==0){
                        control>>wavelength;
                        }
		if(ST(TS,"inname")==0){
			control>>iname;
			}
		if(ST(TS,"outname")==0){
			control>>oname;
			}
                if(ST(TS,"focal_length")==0){
                        control>>F;
                        }
                if(ST(TS,"reconsize")==0){
                        control>>N;
                        }
                if(ST(TS,"pixels")==0){
                        control>>nx>>ny;
                        }
                if(ST(TS,"psize")==0){
                        control>>psize;
                        }
                if(ST(TS,"phase")==0){
                        control>>phaselevels;
                        }
		if(ST(TS,"N")==0){
			control>>N;
			}
		if(ST(TS,"autoscale")==0){
			autoscale=TRUE;
			}
		if(ST(TS,"wavelength")==0){
			wavelength=TRUE;
			}
		}
        while(ST(TS,"endoptions")!=0&&i<MAX_OPTIONS&&!control.eof());


	F_dash=F;
	if(autoscale==TRUE){
		float fov,iov;
		cout<<"SIM_HOL: autoscale zooming enabled\n";
		// select minimum array size for the hologram
		if(N==0){
			N=nx>ny?nx:ny;
			cout<<"Minimum array size is N="<<N<<" (nx,ny)=("<<nx<<","<<ny<<")"<<endl;
			i=2;
			while(i<=N)i*=2;
			N=i;
			}
		cout<<"The simulation size is "<<N<<"\n";
		fov=N*psize;
		iov=wavelength*F/psize;
		F_dash=F*fov/iov;
		cout<<"Zooming to "<<F_dash<<"m instead of "<<F<<"m\n";
		if(F!=F_dash){
			F_bar=1/(1/F-1/F_dash);
			}
		else{
			F_bar=0;
			}
		cout<<"Zoom lens has focal length "<<F_bar<<"\n";
		cout<<"Each pixel is now zoomed from "<<psize<<"m to the equivalent of "<<psize*F/F_dash<<"m\n";
		}

	here=		(complex<double> *)fftw_malloc(sizeof(complex<double>)*N*N);//new complex<double>[N*N];
	there=		(complex<double> *)fftw_malloc(sizeof(complex<double>)*N*N);//new complex<double>[N*N];
	hologram=	new char [nx*ny];
	if(here==NULL || there==NULL || hologram ==NULL){
		cout<<"SIM_HOL: ERROR: unable to allocate main arrays for "<<N<<"*"<<N<<" simulation\n";
		exit(3);
		}
	cout<<"Checking for stored wisdom\n"<<flush;
	fftw_import_system_wisdom();
	if(access(WISDOM_FILE,F_OK|R_OK)==0){
		cout<<"Loading stored wisdom\n"<<flush;
		// we have wisdom stored use it
		wisd_file=fopen(WISDOM_FILE,"r");
		fftw_import_wisdom_from_file(wisd_file);
		fclose(wisd_file);
		}
	cout<<"Computing plans\n"<<flush;
        forward=fftw_plan_dft_2d(N,N,reinterpret_cast<fftw_complex*>(here),reinterpret_cast<fftw_complex*>(there),+1,FFTW_MEASURE);
        backward=fftw_plan_dft_2d(N,N,reinterpret_cast<fftw_complex*>(there),reinterpret_cast<fftw_complex*>(here),-1,FFTW_MEASURE);
	cout<<"Saving wisdom\n"<<flush;
	wisd_file=fopen(WISDOM_FILE,"w");
	fftw_export_wisdom_to_file(wisd_file);
	fclose(wisd_file);
	cout<<"Saving wisdom\n"<<flush;


	// load the data
	cout<<"Reading input file "<<iname<<endl;
	fo=fopen(iname,"r");
	if(fo==NULL){
		cout<<"SIM_HOL: ERROR: Unable to read "<<iname<<"\n";
		exit(3);
		}
	fread((void *)hologram,sizeof(char),nx*ny,fo);
	fclose(fo);
	cout<<"Converting .hol format to complex array \n";
	hol_to_sim(hologram,nx,ny,psize,here,N,F_bar,wavelength,phaselevels);
	cout<<"Performing propagation (1,"<<flush;
	fftw_execute(forward); // goto angular spectrum
	cout<<"2"<<flush;
	prop2d(there,F_dash,psize,wavelength,N);
	cout<<"3)"<<flush;
	fftw_execute(backward); // return to wavefront
	cout<<" done.\n";
	cout<<"Writing out recon file"<<oname<<"\n";
	fo=fopen(oname,"w");
	fwrite((void *)here,sizeof(complex <double>),N*N,fo);
	fclose(fo);	
return 0;
}




void prop2d(complex<double> *data, float F, float px, float w, int N){ 
double A,B,C;
int M,L,i,ii,j,jj,k;
	M=N/2;
	L=M+1;
        A=2*M_PI*F/w;  
        B=w*w/(N*N*px*px);
	// code fix, used to assume large int to make this work (int)(1/B), floor is correct
        C=floor(1/B);

// cribbed from sim_aberration2d because the maths was better than sim_ap and faster
// based of the cache maths principal, recycling the though stuff in each quad.
// more memory access but 1/4 of the trig work (well almost...)
// is much more efficient if (as anywhere else) you recycle the cache many times.
complex <double> *cache;

	// allocate the cache space
	cache =new complex <double> [(M+1)*(M+1)];
	//clear the cache (could be more efficent you know)
        for(i=0;i<L*L;i++){
                cache[i]=complex<float>(0.0,0.0);
                }
	// calc the cache and the 1st quadrant
        // for 1st quad go 0-N/2+1 for others goto 1-N/2 since N/2+1 will be the same either way you look at it
	cout<<"(a,"<<flush;
        for(i=0;(i<L)&&((ii=i*i)<C);i++){
                for(j=0;(j<L)&&((k=ii+j*j)<C);j++){
                        cache[j+L*i]=(complex <float>)polar(1.0,A*(sqrt(1-(B*k))));
			data[j+N*i]*=cache[j+L*i];
                        }
                }
	// check the bandlimit
        M=i;
        if(M!=L){
                cout<<"INFORMATION: Bandlimiting the propagation to "<<M<<" instead of "<<L<<"\n";
                }
	cout<<"b,"<<flush;
        for(i=1,ii=N-1;i<L;i++,ii--){
                for(j=0,jj=N-1;j<M;j++,jj--){
                        data[j+ii*N]*=cache[j+i*L];
                        }
                }
	cout<<"c,"<<flush;
        for(i=0,ii=N-1;i<L;i++,ii--){
                for(j=1,jj=N-1;j<M;j++,jj--){
                        data[jj+i*N]*=cache[j+i*L];
                        }
                }
	cout<<"d"<<flush;
        for(i=1,ii=N-1;i<L;i++,ii--){
                for(j=1,jj=N-1;j<M;j++,jj--){
                        data[jj+ii*N]*=cache[j+i*L];
                        }
                }
	cout<<"),"<<flush;
        }

inline complex<double> rot(complex<double> temp, double phase){
        //rotates a complex<double> by phase radians
	return (polar(sqrt(norm(temp)), arg(temp) + phase));
	}
 
void hol_to_sim(char *hologram, int nx, int ny, float psize, complex <double> * here, int N, float R_curve, float wavelength, int phaselevels){
int x,y,hx,hy;
long hpos,pointer;
float fac;
	hx=nx/2;
	hy=ny/2;
	fac=2*PI/phaselevels;
        for (hpos=0, y=-hy; y< hy; y++){
		for (pointer=N/2-hx+(N/2+y)*N, x=-hx; x<hx; x++, hpos++, pointer++) {
			//position the pointer correctly
			if(hologram[hpos]>=0){
				here[pointer]=polar(1.0,fac*(double)hologram[hpos]);
				}
			else{
				here[pointer]=complex<double>(0,0);
				}
			}
		}
	if(R_curve!=0){
		cout<<"Zooming the wavefront\n";
		add_wavefront(here,nx,ny,N,R_curve,psize,wavelength);
		}
	else{
		cout<<"No zoom\n";
		}

	}
 
void add_wavefront(complex<double> *data, int xsize, int ysize, int N, float R_curve, float psize, float wavelength){
//adds a spherical wavefront to the hologram

	int	sign=1;//direction of curvature
	int m,x,y;
	long pointer;
	float	phase,fac1,fac2;
//only add wavefront to active part of hologram (to add to all xsize and ysize=N)
	if(R_curve<0)sign=-1;
        xsize /= 2;
        ysize /= 2;
        m = N / 2;
	fac1=2*PI*sign/wavelength;
	fac2=psize*psize;
	R_curve=R_curve*R_curve;
        for (y = - ysize; y <  ysize; y++) {
                for (x = - xsize; x <  xsize; x++) {
			//posistion the pointer correctly
                        pointer = (m+x) + (y+m) * N;
			phase=fac1*sqrt(R_curve+fac2*(x*x+y*y));
                        data[pointer] = rot(data[pointer],phase);
                	}
        	}
	} 


void normalizefft(complex<double> *data, unsigned long *nn, int ndim){
	long i,N;
	double scale;
	N=1;
	for(i=1;i<=ndim;i++){
		N*=nn[i];
		}
	scale=1.0/sqrt((double)N);
	for(i=0;i<N;i++){
		data[i]*=scale;
		}
	return;
	}

